# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Implementation of `copyTransactionsFrom`.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import logging
import tempfile

from ZODB.Connection import TransactionMetaData
from ZODB.loglevels import TRACE
from ZODB.blob import is_blob_record
from ZODB.utils import cp as copy_blob
from ZODB.utils import readable_tid_repr
from ZODB.POSException import POSKeyError
from ZODB.interfaces import IStorageCurrentRecordIteration as IRecordIter

from relstorage._compat import perf_counter
from relstorage._util import byte_display

logger = logging.getLogger(__name__)

def close(it):
    try:
        c = it.close
    except AttributeError: # pragma: no cover
        pass
    else:
        c()

class Copy(object):

    __slots__ = (
        'blobhelper',
        'tpc',
        'restore',
    )

    def __init__(self, blobhelper, tpc, restore):
        self.blobhelper = blobhelper
        # In practice, both *tpc* and *restore* are the
        # RelStorage instance.
        self.tpc = tpc
        self.restore = restore

    def copyTransactionsFrom(self, other):
        # Just the interface, not the attribute, in case we have a
        # partial proxy.
        other_has_record_iternext = IRecordIter.providedBy(other)

        copier_factory = _HistoryFreeCopier
        if self.tpc.keep_history or not other_has_record_iternext:
            copier_factory = _HistoryPreservingCopier

        logger.info(
            "Copying transactions to %s "
            "from %s (supports IStorageCurrentRecordIteration? %s) "
            "using %s",
            self.tpc,
            other,
            other_has_record_iternext,
            copier_factory,
        )
        copier = copier_factory(other, self.blobhelper, self.tpc, self.restore)

        try:
            logger.info("Counting the %s to copy.", copier.units)
            num_txns = len(copier)
            logger.info("Copying %d %s%s", num_txns, copier.units, copier.initial_log_suffix)

            progress = copier.ProgressLogger(num_txns, copier)
            copier.copy(progress)
        finally:
            copier.close()

        now = perf_counter()
        logger.info(
            "Copied transactions: %s",
            progress.display_at(now))


class _AbstractCopier(object):

    units = 'transactions'
    initial_log_suffix = ''
    unit_abbrev_display = 'TX'

    __slots__ = (
        'storage',
        'restore',
        'blobhelper',
        'tpc',
        'temp_blobs_to_rm',
        'total_count',
    )

    def __init__(self, storage, blobhelper, tpc, restore):
        self.storage = storage
        self.restore = restore
        self.blobhelper = blobhelper
        self.tpc = tpc
        self.temp_blobs_to_rm = []
        self.total_count = None

    def before_major_log(self):
        raise NotImplementedError

    def __len__(self):
        if self.total_count is None:
            self.total_count = self._compute_total_count()
        return self.total_count

    def _compute_total_count(self):
        raise NotImplementedError

    def copy(self, progress):
        # type: (_ProgressLogger) -> None
        raise NotImplementedError

    @property
    def ProgressLogger(self):
        return _ProgressLogger

    def clean_temp_blobs(self):
        num_blobs = len(self.temp_blobs_to_rm)
        if num_blobs:
            for tmp_blob in self.temp_blobs_to_rm:
                logger.log(TRACE, "Removing temporary blob file %s", tmp_blob)
                try:
                    os.unlink(tmp_blob)
                except OSError:
                    pass
            del self.temp_blobs_to_rm[:]
        return num_blobs

    def close(self):
        self.clean_temp_blobs()
        self.storage = None

    def restore_one(self, active_txn_meta,
                    oid, tid, data):

        # The signature for both ``restore`` and ``restoreBlob``
        # is:
        #
        #   (oid, serial, data, (blobfilename|prev_txn), version, txn)
        #
        # Where ``txn`` is the TransactionMetaData object
        # originally passed to ``tpc_begin``. It is only used to
        # check that the same object has been passed.
        #
        # ``prev_txn`` is not used but would come from ``record.data_txn``

        txn_data_size = len(data) if data else 0

        blobfile = None
        if is_blob_record(data):
            try:
                blobfile = self.storage.openCommittedBlobFile(
                    oid, tid)
            except POSKeyError: # pragma: no cover
                logger.exception("Failed to open blob to copy")

        # We may not be able to read the data after this.
        data = self.restore._crs_transform_record_data(data)
        if blobfile is not None:
            fd, name = tempfile.mkstemp(
                suffix='.tmp',
                dir=self.blobhelper.temporaryDirectory()
            )
            self.temp_blobs_to_rm.append(name)
            logger.log(
                TRACE,
                "Copying %s to temporary blob file %s for upload",
                blobfile, name)

            with os.fdopen(fd, 'wb') as target:
                # If we don't get the length, ``copy_blob`` will.
                old_pos = blobfile.tell()
                blobfile.seek(0, 2)
                length = blobfile.tell()
                blobfile.seek(old_pos)

                copy_blob(blobfile, target, length)
                txn_data_size += length
            blobfile.close()
            self.restore.restoreBlob(oid, tid, data,
                                     name, None, active_txn_meta)
        else:
            self.restore.restore(oid, tid, data,
                                 '', None, active_txn_meta)

        return txn_data_size, blobfile is not None


class _HistoryPreservingCopier(_AbstractCopier):

    __slots__ = (
        'storage_it',
    )

    def __init__(self, storage, blobhelper, tpc, restore):
        super(_HistoryPreservingCopier, self).__init__(storage, blobhelper, tpc, restore)
        self.storage_it = storage.iterator()

    def _compute_total_count(self):
        try:
            num_txns = len(self.storage_it)
            if num_txns == 0:
                # Hmm, that shouldn't really be right, should it?
                # Try the other path.
                raise TypeError()
        except TypeError:
            logger.warning("Iterator %s doesn't support len(); counting manually", self.storage_it)
            num_txns = 0
            for _ in self.storage_it:
                num_txns += 1
            close(self.storage_it)
            self.storage_it = self.storage.iterator()

        return num_txns

    def before_major_log(self):
        pass

    def close(self):
        close(self.storage_it)
        self.storage_it = None
        super(_HistoryPreservingCopier, self).close()

    def copy(self, progress):
        # type: (_ProgressLogger) -> None
        for trans in self.storage_it:
            begin = perf_counter()
            num_txn_records, txn_data_size, num_blobs = self(trans)
            now = perf_counter()
            progress.copied_one(now, now - begin, trans, num_txn_records, txn_data_size, num_blobs)

    def __call__(self, trans):
        # Originally adapted from ZODB.blob.BlobStorageMixin
        tpc = self.tpc
        num_txn_records = 0
        txn_data_size = 0

        tpc.tpc_begin(trans, trans.tid, trans.status)
        for record in trans:
            num_txn_records += 1
            record_size, _was_blob = self.restore_one(
                trans,
                record.oid,
                record.tid,
                record.data
            )
            txn_data_size += record_size
        tpc.tpc_vote(trans)
        tpc.tpc_finish(trans)

        num_blobs = self.clean_temp_blobs()

        return num_txn_records, txn_data_size, num_blobs


class _RecordIternextIterator(object):

    __slots__ = (
        'storage',
        'cookie',
    )

    def __init__(self, storage):
        self.storage = storage
        self.cookie = self

    def __iter__(self): # pragma: no cover
        return self

    def __next__(self):
        if self.cookie is None:
            raise StopIteration

        if self.cookie is self:
            # First time in.
            self.cookie = None
        try:
            result = self.storage.record_iternext(self.cookie)
        except ValueError:  # pragma: no cover
            # FileStorage can raise this if the underlying storage
            # is completely empty.
            # See https://github.com/zopefoundation/ZODB/issues/330
            if self.cookie is None:
                raise StopIteration
            raise # pragma: no cover

        oid, tid, state, self.cookie = result
        return oid, tid, state

    next = __next__ # Py2


class _HistoryFreeTransactionMetaData(TransactionMetaData):
    num_records = 0
    num_records_since_last_log = 0
    num_blobs_since_last_log = 0
    record_size_since_last_log = 0
    first_oid = None
    last_oid = None

    def reset_interval(self):
        self.num_records_since_last_log = 0
        self.num_blobs_since_last_log = 0
        self.record_size_since_last_log = 0


class _HistoryFreeCopier(_AbstractCopier):
    # Issues with iternext:
    # - No way to specify a starting transaction (FileStorage and RelStorage
    #   iterate by OID and can specify a starting OID).
    #   So this silently breaks the ``incremental`` mode of zodbconvert;
    #   Things still work, we just copy all objects from the beginning, and
    #   if we copied some, stopped, then start again, and in the meantime the source
    #   storage was packed, we could wind up with extra objects. That's the case
    #   anyway, though. The documentation has been set up to warn
    #   about that. TODO: Figure out an incremental way.
    # - Doesn't support len(); Rather than iterating manually to find it,
    #   and thus downloading all the object states, or copying them into a local FileStorage,
    #   (which would use lots of temp space) we ask for the estimated size from the
    #   storage.
    #
    # Issues copying from history-preserving to history-free
    # using the regular iterator():
    #
    # - We could copy and discard the state for a single object
    #   many times. Doing way too much work.

    units = 'objects'
    unit_abbrev_display = 'obj'
    initial_log_suffix = '. CAUTION: This number is approximate.'

    __slots__ = (
        'trans_meta',
    )

    def __init__(self, *args, **kwargs):
        _AbstractCopier.__init__(self, *args, **kwargs)
        self.trans_meta = None

    def _compute_total_count(self):
        return len(self.storage)

    def __iter__(self):
        return _RecordIternextIterator(self.storage)

    def copy(self, progress):
        # type: (_HistoryFreeProgressLogger) -> None
        for oid, tid, state in self:
            begin = perf_counter()
            if self.trans_meta is None:
                self.trans_meta = _HistoryFreeTransactionMetaData()
                self.trans_meta.first_oid = oid
                self.tpc.tpc_begin(self.trans_meta, tid)

            self.trans_meta.last_oid = oid
            record_size, was_blob = self.restore_one(
                self.trans_meta,
                oid,
                tid,
                state
            )
            self.trans_meta.num_records += 1
            self.trans_meta.num_records_since_last_log += 1
            self.trans_meta.num_blobs_since_last_log += was_blob
            self.trans_meta.record_size_since_last_log += record_size

            now = perf_counter()

            progress.copied_one(now, now - begin, self.trans_meta, 1, record_size,
                                was_blob)

        # Perform the final commit if needed.
        self.before_major_log()

    @property
    def ProgressLogger(self):
        return _HistoryFreeProgressLogger

    def before_major_log(self):
        # Integrate committing with the major log intervals. This is because the
        # major log interval is time-based and tunable. And, up to a point, committing
        # a larger batch provides throughput improvements.
        if self.trans_meta is not None:
            logger.debug('Committing object batch count=%d', self.trans_meta.num_records)
            trans_meta = self.trans_meta
            self.trans_meta = None
            self.tpc.tpc_vote(trans_meta)
            self.tpc.tpc_finish(trans_meta)
            self.clean_temp_blobs()


class _ProgressLogger(object):

    # Time in seconds between major progress logging.
    # (minor progress logging occurs every ``minor_log_count`` commits)
    log_interval = 60

    # Number of units to copy before checking if we should perform a major
    # log.
    log_count = 100

    # Number of units to copy before checking if we should perform a minor log.
    minor_log_count = 25

    minor_log_interval = 15

    minor_log_tx_record_count = 100
    minor_log_tx_size = 500 * 1024
    minor_log_copy_time_threshold = 1.0

    debug_enabled = False

    class _IntervalStats(object):
        __slots__ = (
            'begin_time',
            'units_copied',
            'total_size',
        )

        def __init__(self, begin_time):
            self.begin_time = begin_time
            self.units_copied = 0
            self.total_size = 0

        def display_at(self, now, total_num_units, unit_abbrev_display,
                       include_elapsed=False):
            pct_complete = '%1.2f%%' % ((
                (self.units_copied * 100.0 if total_num_units else 1)
                /
                (total_num_units or 1)
            ))

            elapsed_total = now - self.begin_time
            rate_mb = rate_units = 0.0
            if elapsed_total:
                rate_mb = self.total_size / elapsed_total
                rate_units = self.units_copied / elapsed_total

            rate_mb_str = byte_display(rate_mb)
            rate_unit_str = '%1.2f' % rate_units

            result = "%d/%d,%7s, %6s/s %6s %s/s, %s" % (
                self.units_copied, total_num_units, pct_complete,
                rate_mb_str, rate_unit_str, unit_abbrev_display,
                byte_display(self.total_size),
            )
            if include_elapsed:
                result += ' %4.1f minutes' % (elapsed_total / 60.0)
            return result

    def __init__(self, num_txns, copier):
        self.num_txns = num_txns # type: int
        self.copier = copier # type: _AbstractCopier
        begin_time = perf_counter()
        self._entire_stats = self._IntervalStats(begin_time)
        self._interval_stats = self._IntervalStats(begin_time)

        self.log_at = begin_time + self.log_interval
        self.minor_log_at = begin_time + self.minor_log_interval
        self.debug_enabled = logger.isEnabledFor(logging.DEBUG) or type(self).debug_enabled

    def display_at(self, now):
        return self._entire_stats.display_at(now, self.num_txns, True)

    def copied_one(self, now, copy_duration, trans,
                   num_txn_records, txn_byte_size, num_txn_blobs):
        # type: (float, float, Any, int, int, int)
        entire_stats = self._entire_stats
        interval_stats = self._interval_stats

        entire_stats.units_copied += 1
        interval_stats.units_copied += 1
        total_units_copied = self._entire_stats.units_copied

        entire_stats.total_size += txn_byte_size
        interval_stats.total_size += txn_byte_size

        if self.debug_enabled and self._should_minor_log(
                now,
                total_units_copied,
                txn_byte_size,
                num_txn_records,
                copy_duration
        ):
            self.minor_log_at = now + self.minor_log_interval
            self.do_minor_log(trans, num_txn_records, txn_byte_size, num_txn_blobs, copy_duration)


        if self._should_major_log(now, total_units_copied):
            self.copier.before_major_log()
            now = perf_counter()
            self.log_at = now + self.log_interval
            self.__major_log(
                now,
                self.transaction_display(trans, num_txn_records, txn_byte_size, num_txn_blobs))
            self._interval_stats = self._IntervalStats(now)

    def _should_major_log(self, now, total_units_copied):
        return total_units_copied % self.log_count and now >= self.log_at

    def _should_minor_log(self, now, total_units_copied, txn_byte_size, num_txn_records,
                          copy_duration): # pragma: no cover
        # This is mocked out by test_zodbconvert so we always try to log.
        if (total_units_copied % self.minor_log_count == 0 and now >= self.minor_log_at):
            return True
        if txn_byte_size >= self.minor_log_tx_size:
            return True
        if num_txn_records >= self.minor_log_tx_record_count:
            return True
        if copy_duration >= self.minor_log_copy_time_threshold:
            return True
        return False

    def do_minor_log(self, trans, num_txn_records, txn_byte_size, num_txn_blobs, copy_duration):
        logger.debug(
            "Copied %s in %1.4fs",
            self.transaction_display(trans, num_txn_records,
                                       txn_byte_size, num_txn_blobs),
            copy_duration
        )

    def __major_log(self, now, transaction_display):
        logger.info(
            "Copied %s | %60s | (%s)",
            transaction_display,
            self._interval_stats.display_at(now, self.num_txns, self.copier.unit_abbrev_display),
            self._entire_stats.display_at(now, self.num_txns,
                                          self.copier.unit_abbrev_display,
                                          include_elapsed=True)
        )

    def transaction_display(self, trans, num_txn_records, txn_byte_size, num_txn_blobs):
        return 'transaction %s <%4d records, %3d blobs, %9s>' % (
            readable_tid_repr(trans.tid),
            num_txn_records, num_txn_blobs, byte_display(txn_byte_size)
        )


class _HistoryFreeProgressLogger(_ProgressLogger):
    minor_log_tx_size = 1024 * 1024 # Only log if a single object is larger than 1MB

    def do_minor_log(self, trans, _num_txn_records, txn_byte_size, _num_txn_blobs, _copy_duration):
        logger.debug("Copied %d objects of size %s (%d blobs) (Most recent oid=%s size=%s)",
                     trans.num_records_since_last_log,
                     byte_display(trans.record_size_since_last_log),
                     trans.num_blobs_since_last_log,
                     trans.last_oid,
                     byte_display(txn_byte_size))
        trans.reset_interval()

    def transaction_display(self, trans, _num_txn_records, _txn_byte_size, _num_txn_blobs):
        return "object range %s -> %s" % (
            trans.first_oid,
            trans.last_oid,
        )



# -*- coding: utf-8 -*-
"""
Tests for connections.py

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.tests import MockConnectionManager

from ..interfaces import IManagedStoreConnection
from ..interfaces import IManagedLoadConnection
from ..interfaces import IManagedDBConnection
from ..connections import StoreConnection
from ..connections import LoadConnection
from ..connections import ClosedConnection
from ..connections import StoreConnectionPool

class TestConnectionCommon(TestCase):

    klass = StoreConnection
    iface = IManagedStoreConnection

    def _makeArgument(self):
        return MockConnectionManager(clean_rollback=False)

    def _makeOne(self):
        return self.klass(self._makeArgument())

    def test_provides(self):
        assert_that(self._makeOne(), validly_provides(self.iface))


class TestConnection(TestConnectionCommon):

    def test_restart_and_call_does_not_activate(self):
        manager = self._makeOne()
        self.assertFalse(manager.active)
        self.assertFalse(manager)

        def on_first_use(_conn, _cursor):
            self.fail('Should not call')

        cc = []

        def on_opened(conn, cursor):
            self.assertIsNotNone(conn)
            self.assertIsNotNone(cursor)
            cc.append(conn)
            cc.append(cursor)

        manager.on_first_use = on_first_use
        manager.on_opened = on_opened

        manager.restart_and_call(lambda _conn, _cursor: None)

        # If we try to access the cursor, it gets called, though.
        with self.assertRaises(AssertionError):
            getattr(manager, 'cursor')

        del manager.on_opened
        del manager.on_first_use
        self.assertEqual(cc, [manager.connection, manager.cursor])

    def test_restart_and_call_opens(self):
        manager = self._makeOne()
        opened = [0]
        def on_opened(_conn, _cursor):
            opened[0] += 1
        manager.on_opened = on_opened
        manager.restart_and_call(lambda _conn, _cursor: None)

        self.assertEqual(opened, [1])

    def test_restart_and_call_rollback(self):
        manager = self._makeOne()
        opened = [0]
        def on_rolledback(_conn, _cursor):
            opened[0] += 1
        manager.on_rolledback = on_rolledback

        manager.restart_and_call(lambda _conn, _cursor: None)
        # That was just an open, so no rollback
        self.assertEqual(opened, [0])

        # But second time is the charm
        manager.restart_and_call(lambda _conn, _cursor: None)
        self.assertEqual(opened, [1])

    def test_drop_activates(self, meth='drop'):
        count = [0]

        def on_first_use(_conn, _cursor):
            count[0] += 1

        manager = self._makeOne()
        manager.on_first_use = on_first_use

        getattr(manager, 'cursor')
        self.assertEqual(count, [1])

        getattr(manager, meth)()
        self.assertEqual(count, [1])

        getattr(manager, 'cursor')
        self.assertEqual(count, [2])

    def test_rollback_activates(self):
        self.test_drop_activates('rollback_quietly')

    def test_call_reconnect(self):
        manager = self._makeOne()
        manager.open_if_needed()
        called = []
        def f(_conn, _cur, fresh):
            called.append(fresh)
            if not fresh:
                raise manager.connmanager.driver.disconnected_exceptions[0]

        manager.call(f, can_reconnect=True)

        self.assertEqual(called, [False, True])

class TestLoadConnection(TestConnection):
    klass = LoadConnection
    iface = IManagedLoadConnection

class TestClosedConnection(TestConnectionCommon):
    klass = ClosedConnection
    iface = IManagedDBConnection



class TestStoreConnectionPool(TestCase):

    def test_borrowing_doesnt_leak_if_creating_raises_exception(self):
        connmanager = MockConnectionManager()

        class MyException(Exception):
            pass

        class StoreConnectionNoBegin(StoreConnection):
            dropped = 0

            def drop(self):
                type(self).dropped += 1
                StoreConnection.drop(self)

            def begin(self):
                raise MyException

        def check(factory, ex):
            pool = StoreConnectionPool(connmanager)

            pool._factory = factory

            with self.assertRaises(ex):
                with pool.borrowing():
                    pass
            self.assertLength(pool._connections, 0)

        check(StoreConnectionNoBegin, MyException)
        self.assertEqual(StoreConnectionNoBegin.dropped, 1)

        class MyBeginException(Exception):
            pass

        def factory(_):
            raise MyBeginException

        check(factory, MyBeginException)

    def test_shrinking_removes_at_beginning(self):
        connmanager = MockConnectionManager(clean_rollback=True)
        connmanager.begin = lambda *_args: None

        pool = StoreConnectionPool(connmanager)

        borrowed = [pool.new_instance().borrow() for _ in range(10)]

        for c in borrowed:
            pool.replace(c)

        self.assertEqual(borrowed, pool._connections)

        pool.MAX_STORE_CONNECTIONS_IN_POOL = 5
        remaining = pool._connections[5:]
        pool._shrink()
        self.assertEqual(remaining, pool._connections)

    def test_returning_dropped_does_not_add_to_pool(self):
        connmanager = MockConnectionManager(clean_rollback=True)
        connmanager.begin = lambda *_args: None

        pool = StoreConnectionPool(connmanager)
        conn = pool.borrow()
        assert conn.connection is not None
        conn.drop()
        assert conn.connection is None
        pool.replace(conn)
        self.assertEqual(0, pool.pooled_connection_count)

    def _makeSemiConnManager(self):
        """
        Return a real conn manager, with minimal mocking.
        """
        from ..connmanager import AbstractConnectionManager
        from relstorage.tests import MockOptions
        from relstorage.tests import MockDriver
        from relstorage.tests import MockConnection

        class ConnManager(AbstractConnectionManager):
            def __init__(self, options=None, driver=None):
                super(ConnManager, self).__init__(options or MockOptions(),
                                                  driver or MockDriver())

                self.begin_count = 0
                self.restart_count = 0

            def restart_store(self, *_args, **_kwargs):
                self.restart_count += 1
                super(ConnManager, self).restart_store(*_args, **_kwargs)

            def begin(self, conn, cursor):
                self.begin_count += 1
                super(ConnManager, self).begin(conn, cursor)

            def open(self, *_args, **_kwargs):
                conn = MockConnection()
                return conn, conn.cursor()

            def _do_open_for_load(self, *_args, **_kwargs):
                raise NotImplementedError
        return ConnManager

    def test_borrow_calls_on_opened_restart(self):
        on_opened_count = []
        def do_on_store_opened(_cursor, restart=None):
            on_opened_count.append(restart)

        connmanager = self._makeSemiConnManager()()
        connmanager.add_on_store_opened(do_on_store_opened)

        pool = StoreConnectionPool(connmanager)

        # restart isn't called to borrow the connection
        store_conn = pool.borrow()
        self.assertEqual(1, connmanager.begin_count)
        self.assertEqual(0, connmanager.restart_count)
        # nor to open it,
        store_conn.open_if_needed()
        # nor to get the cursor, making it active.
        getattr(store_conn, 'cursor')
        self.assertTrue(store_conn)
        self.assertTrue(store_conn.active)
        self.assertEqual(1, connmanager.begin_count)
        self.assertEqual(0, connmanager.restart_count)
        self.assertEqual(on_opened_count, [False])

        pool.replace(store_conn)
        self.assertFalse(store_conn)
        self.assertFalse(store_conn.active)

        # Begin is called on borrowing,
        # and so is restart
        store_conn2 = pool.borrow()
        self.assertIs(store_conn, store_conn2)
        self.assertEqual(2, connmanager.begin_count)
        self.assertEqual(1, connmanager.restart_count)
        self.assertEqual(on_opened_count, [False, True])

    def test_restart_handles_replica_closed(self):
        class MockReplicaSelector(object):
            def current(self):
                return 1

        connmanager = self._makeSemiConnManager()()
        assert hasattr(connmanager, 'replica_selector')
        connmanager.replica_selector = MockReplicaSelector()
        pool = StoreConnectionPool(connmanager)

        store_conn = pool.borrow()
        orig_raw_conn = store_conn.connection

        store_conn.open_if_needed()
        getattr(store_conn, 'cursor')
        pool.replace(store_conn)

        store_conn2 = pool.borrow()
        # Same object,
        self.assertIs(store_conn, store_conn2)
        del store_conn
        # But different internals
        self.assertIsNot(store_conn2.connection, orig_raw_conn)



# -*- coding: utf-8 -*-
"""
Tests for temporary_storage.py

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

class TestTemporaryStorage(unittest.TestCase):

    def _makeOne(self):
        from ..temporary_storage import TemporaryStorage
        return TemporaryStorage()

    def setUp(self):
        super(TestTemporaryStorage, self).setUp()
        from .. import temporary_storage
        temporary_storage.id = lambda _: 0xDEADBEEF

    def tearDown(self):
        from .. import temporary_storage
        del temporary_storage.id
        super(TestTemporaryStorage, self).tearDown()

    _EMPTY_STR = '<TPCTemporaryStorage at 0xdeadbeef count=0 bytes=0>'

    def test_empty_str(self):
        temp = self._makeOne()
        s = str(temp)
        self.assertEqual(
            s,
            self._EMPTY_STR
        )

    def test_closed_str(self):
        temp = self._makeOne()
        temp.close()
        s = str(temp)
        self.assertEqual(s, self._EMPTY_STR)

    def test_str_with_data(self):
        from textwrap import dedent
        self.maxDiff = None
        temp = self._makeOne()
        temp.store_temp(6547, b'defghijkl', 23)
        temp.store_temp(1, b'abc')
        temp.store_temp(2, b'def', 42)

        s = str(temp)

        self.assertEqual(
            s,
            dedent("""\
            <TPCTemporaryStorage at 0xdeadbeef count=3 bytes=15>
            ====================================================
            | OID            | Length         | Previous TID  |
            ====================================================
                          1  |              3 |              0
                          2  |              3 |             42
                       6547  |              9 |             23
            """
                   )
        )



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

from collections import namedtuple

from zope.interface import implementer

from ZODB.utils import p64 as int64_to_8bytes

from relstorage._compat import TidList
from relstorage._util import Lazy

from ._util import DatabaseHelpersMixin
from .interfaces import IDatabaseIterator
from .schema import Schema
from .sql import it

class DatabaseIterator(DatabaseHelpersMixin):
    """
    Abstract base class for database iteration.
    """

    def __init__(self, database_driver):
        """
        :param database_driver: Necessary to bind queries correctly.
        """
        self.driver = database_driver

    @Lazy
    def _as_state(self):
        return self.driver.binary_column_as_state_type

    @Lazy
    def _as_bytes(self):
        return self.driver.binary_column_as_bytes

    _iter_objects_query = Schema.object_state.select(
        it.c.zoid,
        it.c.state
    ).where(
        it.c.tid == it.bindparam('tid')
    ).order_by(
        it.c.zoid
    )

    def iter_objects(self, cursor, tid):
        """Iterate over object states in a transaction.

        Yields ``(oid, state)`` for each object in the transaction.
        """
        self._iter_objects_query.execute(cursor, {'tid': tid})
        as_state = self._as_state
        for oid, state in cursor:
            state = as_state(state) # pylint:disable=too-many-function-args
            yield oid, state

    _iter_current_records = Schema.all_current_object_state.select(
        it.c.zoid,
        it.c.tid,
        it.c.state
    ).where(
        it.c.zoid >= it.bindparam('start_oid')
    ).order_by(
        it.c.zoid
    )

    def iter_current_records(self, cursor, start_oid_int=0):
        """
        Cause the *cursor* (which should be a server-side cursor)
        to execute a query that will iterate over
        ``(oid_int, tid_int, state_bytes)`` values for all the current objects.

        Each current object is returned only once, at the transaction most recently
        committed for it.

        Objects are iterated or order of their OID for compatibility with FileStorage.

        Returns a generator.
        """
        self._iter_current_records.execute(cursor, {'start_oid': start_oid_int})
        i_b = int64_to_8bytes
        s = self._as_state
        for oid_int, tid_int, state_bytes in cursor:
            yield i_b(oid_int), i_b(tid_int), s(state_bytes) # pylint:disable=too-many-function-args


class _HistoryPreservingTransactionRecord(namedtuple(
        '_HistoryPreservingTransactionRecord',
        ('tid_int', 'username', 'description', 'extension', 'packed')
)):
    __slots__ = ()

    @property
    def pickle_size(self):
        return self.packed


@implementer(IDatabaseIterator)
class HistoryPreservingDatabaseIterator(DatabaseIterator):

    keep_history = True

    def _transaction_iterator(self, cursor):
        """
        Iterate over a list of transactions returned from the database.

        Each row is ``(tid, username, description, extension, X)``
        """
        # Iterating the cursor itself in a generator is not safe if
        # the cursor doesn't actually buffer all the rows *anyway*. If
        # we break from the iterating loop before exhausting all the
        # rows, a subsequent query or close operation can lead to
        # things like MySQL Connector/Python raising
        # InternalError(unread results)
        # Because we have it all in memory anyway, there's not much point in
        # making this a generator.

        # Although the transaction interface for username and description are
        # defined as strings, this layer works with bytes. The ZODB layer
        # does the conversion.
        as_bytes = self._as_bytes
        # pylint:disable=too-many-function-args
        return [
            _HistoryPreservingTransactionRecord(
                tid,
                as_bytes(username),
                as_bytes(description),
                as_bytes(ext),
                packed
            )
            for (tid, username, description, ext, packed)
            in cursor
        ]

    _iter_transactions_query = Schema.transaction.select(
        it.c.tid, it.c.username, it.c.description, it.c.extension, 0
    ).where(
        it.c.packed == False # pylint:disable=singleton-comparison
    ).and_(
        it.c.tid != 0
    ).order_by(
        it.c.tid, 'DESC'
    )

    def iter_transactions(self, cursor):
        """Iterate over the transaction log, newest first.

        Skips packed transactions.
        Yields (tid, username, description, extension) for each transaction.
        """
        self._iter_transactions_query.execute(cursor)
        return self._transaction_iterator(cursor)

    _iter_transactions_range_query = Schema.transaction.select(
        it.c.tid,
        it.c.username,
        it.c.description,
        it.c.extension,
        it.c.packed,
    ).where(
        it.c.tid >= it.bindparam('min_tid')
    ).and_(
        it.c.tid <= it.bindparam('max_tid')
    ).order_by(
        it.c.tid
    )

    def iter_transactions_range(self, cursor, start=None, stop=None):
        """
        See `IDatabaseIterator`.
        """
        params = {
            'min_tid': start if start else 0,
            'max_tid': stop if stop else self.MAX_TID
        }
        self._iter_transactions_range_query.execute(cursor, params)
        return self._transaction_iterator(cursor)

    _object_exists_query = Schema.current_object.select(
        1
    ).where(
        it.c.zoid == it.bindparam('oid')
    )

    _object_history_query = Schema.transaction.natural_join(
        Schema.object_state
    ).select(
        it.c.tid, it.c.username, it.c.description, it.c.extension,
        Schema.object_state.c.state_size
    ).where(
        it.c.zoid == it.bindparam("oid")
    ).and_(
        it.c.packed == False # pylint:disable=singleton-comparison
    ).order_by(
        it.c.tid, "DESC"
    )

    def iter_object_history(self, cursor, oid):
        """
        See `IDatabaseIterator`
        Raises KeyError if the object does not exist.
        """
        params = {'oid': oid}
        self._object_exists_query.execute(cursor, params)
        if not cursor.fetchall():
            raise KeyError(oid)

        self._object_history_query.execute(cursor, params)
        return self._transaction_iterator(cursor)

class _HistoryFreeTransactionRecord(object):
    __slots__ = ('tid_int',)

    username = b''
    description = b''
    extension = b''
    packed = True

    def __init__(self, tid):
        self.tid_int = tid


class _HistoryFreeObjectHistoryRecord(_HistoryFreeTransactionRecord):
    __slots__ = ('pickle_size',)

    def __init__(self, tid, size):
        _HistoryFreeTransactionRecord.__init__(self, tid)
        self.pickle_size = size


class _HistoryFreeTransactionRange(object):
    # By storing just the int, and materializing the records on demand, we
    # save substantial amounts of memory. For example, 18MM records on PyPy
    # went from about 3.5GB to about 0.5GB
    __slots__ = (
        'tid_ints',
    )

    def __init__(self, tid_ints):
        self.tid_ints = tid_ints

    def __len__(self):
        return len(self.tid_ints)

    def __getitem__(self, ix):
        return _HistoryFreeTransactionRecord(self.tid_ints[ix])

@implementer(IDatabaseIterator)
class HistoryFreeDatabaseIterator(DatabaseIterator):

    keep_history = False

    def iter_transactions(self, cursor):
        """
        This always returns an empty iterable.
        """
        # pylint:disable=unused-argument
        return ()

    _iter_transactions_range_query = Schema.object_state.select(
        it.c.tid,
    ).where(
        it.c.tid >= it.bindparam('min_tid')
    ).and_(
        it.c.tid <= it.bindparam('max_tid')
    ).order_by(
        it.c.tid
    ).distinct()

    def iter_transactions_range(self, cursor, start=None, stop=None):
        """
        See `IDatabaseIterator`.
        """
        params = {
            'min_tid': start if start else 0,
            'max_tid': stop if stop else self.MAX_TID
        }
        self._iter_transactions_range_query.execute(cursor, params)
        return _HistoryFreeTransactionRange(TidList((tid for (tid,) in cursor)))

    _iter_object_history_query = Schema.object_state.select(
        it.c.tid, it.c.state_size
    ).where(
        it.c.zoid == it.bindparam('oid')
    )

    def iter_object_history(self, cursor, oid):
        """
        See `IDatabaseIterator`

        Yields a single row.
        """
        self._iter_object_history_query.execute(cursor, {'oid': oid})
        rows = cursor.fetchall()
        if not rows:
            raise KeyError(oid)
        assert len(rows) == 1
        tid, size = rows[0]
        return [_HistoryFreeObjectHistoryRecord(tid, size)]



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ..batch import RowBatcher

class Sqlite3RowBatcher(RowBatcher):
    # The batch size depends on how many params a statement can
    # have; if we go too big we get OperationalError: too many SQL
    # variables. The default allowed is 999.
    # Note that the multiple-value syntax was added in
    # 3.7.11, 2012-03-20.

    bind_limit = 998

    # sqlite only supports ? as a param.
    delete_placeholder = '?'
    insert_placeholder = '?'



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Batch table row insert/delete support.
"""
from collections import defaultdict
import itertools

from relstorage._compat import iteritems
from relstorage._compat import perf_counter
from relstorage._util import parse_byte_size

from .interfaces import AggregateOperationTimeoutError

# A cache
# {(placeholder, count): "string"}
_row_schemas = {}

class RowBatcher(object):
    """
    Generic row batcher.

    Inserting can use whatever paramater placeholder format
    and parameter format the cursor supports.

    Deleting needs to use ordered parameters. The placeholder
    can be set in the ``delete_placeholder`` attribute.
    """

    perf_counter = perf_counter

    # How many total rows can be sent at once. Also used as
    # ``bind_limit`` if that is 0 or None.
    row_limit = 1024
    # The total number of available bind variables a single statement
    # can use.
    bind_limit = None
    # The default max_allowed_packet in MySQL is 4MB,
    # so the data, including encoding and the rest of the query structure,
    # must be less than that (unless we dynamically query to find out
    # what value we have)
    size_limit = parse_byte_size('2 MB')
    delete_placeholder = '%s'
    insert_placeholder = '%s'
    # For testing, force the delete order to be deterministic
    # when multiple columns are involved
    sorted_deletes = False

    def __init__(self, cursor, row_limit=None,
                 delete_placeholder=None,
                 insert_placeholder=None,
                 bind_limit=None):
        self.cursor = cursor
        if delete_placeholder is not None:
            self.delete_placeholder = delete_placeholder
        if insert_placeholder is not None:
            self.insert_placeholder = insert_placeholder
        if row_limit is not None:
            self.row_limit = row_limit
        if bind_limit is not None:
            self.bind_limit = bind_limit

        # These are cumulative
        self.total_rows_inserted = 0
        self.total_size_inserted = 0
        self.total_rows_deleted = 0

        # These all get reset at each flush()
        self.rows_added = 0
        self.size_added = 0
        self.bind_params_added = 0

        self.deletes = defaultdict(set)   # {(table, columns_tuple): set([(column_value,)])}
        self.inserts = defaultdict(dict)  # {(command, header, row_schema, suffix): {rowkey: [row]}}

    def __repr__(self):
        return "<%s at %x tins=%d tdel=%d prows=%d>" % (
            self.__class__.__name__,
            id(self),
            self.total_rows_inserted,
            self.total_rows_deleted,
            self.rows_added,
        )

    def _flush_if_needed(self):
        """
        Return the number of rows updated.
        """
        if self.rows_added >= self.row_limit:
            return self.flush()
        if self.bind_limit and self.bind_params_added >= self.bind_limit:
            return self.flush()
        if self.size_added >= self.size_limit:
            return self.flush()
        return 0

    def _flush_if_would_exceed_bind(self, addition):
        # The bind limit is a hard limit we cannot exceed.
        # If adding *addition* params would cause us to exceed,
        # flush now.
        if self.bind_limit and self.bind_params_added + addition >= self.bind_limit:
            return self.flush() or True # This should always be at least one, right?
        return 0

    def delete_from(self, table, **kw):
        """
        Returns the number of rows flushed as a result of this operation.
        That can include inserts.
        """
        # XXX: When deleting a lot from a single table, a bulk function
        # might be a lot faster.
        if not kw:
            raise AssertionError("Need at least one column value")
        columns = tuple(sorted(kw))
        key = (table, columns)
        row = tuple(kw[column] for column in columns)
        bind_params_added = len(row) if key not in self.deletes[key] else 0
        count = self._flush_if_would_exceed_bind(bind_params_added)

        self.deletes[key].add(row)
        self.rows_added += 1
        self.bind_params_added += bind_params_added
        count += self._flush_if_needed()
        return count

    def insert_into(self, header, row_schema, row, rowkey, size,
                    command='INSERT', suffix=''):
        key = (command, header, row_schema, suffix)

        bind_params_added = len(row) if rowkey not in self.inserts[key] else 0
        self._flush_if_would_exceed_bind(bind_params_added)

        # If we flushed, self.inserts has started all over.
        self.inserts[key][rowkey] = row  # note that this may replace a row

        self.rows_added += 1
        self.bind_params_added += bind_params_added
        self.size_added += size
        self._flush_if_needed()

    def row_schema_of_length(self, param_count):
        # Use as the *row_schema* parameter to insert_into
        key = (self.insert_placeholder, param_count)
        try:
            return _row_schemas[key]
        except KeyError:
            p = [self.insert_placeholder] * param_count
            p = ', '.join(p)
            _row_schemas[key] = p
            return p

    def select_from(self, columns, table, suffix='', timeout=None, **kw):
        """
        Handles a query of the ``WHERE col IN (?, ?,)`` type::

        ``SELECT columns FROM table WHERE col IN (?, ...)``

        The keyword arguments should be of length 1, containing an
        iterable of the values to check: ``col=(1, 2)`` or in the
        dynamic case ``**{indirect_var: [1, 2]}``::

            batcher.select_from(('zoid', 'tid',), 'object_state',
                                zoid=oids)

        The number of batches needed is determined by the length of
        the iterator divided by this object's ``bind_limit``,
        or, if that's not set, by the ``row_limit``.

        Returns a iterator of matching rows. Matching rows are delivered
        incrementally, so some number of rows may be delivered and then
        an exception is raised.

        :keyword float timeout: If given, provides a number of seconds
           that is the approximate maximum amount of time this method will
           be allowed to take.
        :raises AggregateOperationTimeoutError: If *timeout* is given,
           and the cumulative time taken to query and process
           some subset of batches exceeds *timeout*. This is checked
           after each individual batch.
        """
        command = 'SELECT %s' % (','.join(columns),)

        for cursor, _count in self.__select_like(
                command,
                table,
                suffix,
                timeout,
                kw
        ):
            for row in cursor.fetchall():
                yield row

    def update_set_static(self, update_set, timeout=None,
                          batch_done_callback=lambda total_count: None,
                          **kw):
        """
        As for :meth:`select_from`, but the first parameter is
        the complete static UPDATE statement. It must be uppercase,
        and startwith "UPDATE".

        Rows are net expected to be returned, so the cursor is completely consumed between
        batches.
        """
        # We actually just consume the iteration of the cursor itself;
        # we can't consume the cursor. Some drivers (pg8000) throw ProgrammingError
        # if we try to iterate the cursor that has no rows.
        total_count = 0
        for _cursor, total_count in self.__select_like(
                update_set,
                None, # table not used
                '', # suffix not used,
                timeout,
                kw
        ):
            batch_done_callback(total_count)

        return total_count

    def __select_like(self, command, table, suffix, timeout, kw):
        assert len(kw) == 1, kw
        # filter_values may be a generic iterable or even a generator;
        # be sure not to materialize the whole thing at any point. Never
        # more than a chunk_size at a time.
        filter_column, filter_values = kw.popitem()
        filter_values = iter(filter_values)

        chunk_size = self.bind_limit or self.row_limit
        chunk_size -= 1

        begin = self.perf_counter() if timeout else None

        count = 0
        for head in filter_values:
            filter_subset = list(itertools.islice(filter_values, chunk_size))
            filter_subset.append(head)

            descriptor = [[(table, (filter_column,)), filter_subset]]

            count += self._do_batch(command, descriptor, rows_need_flattened=False, suffix=suffix)

            yield self.cursor, count

            if timeout and self.perf_counter() - begin >= timeout:
                # TODO: It'd be nice not to do this if we had no more
                # batches to do.
                raise AggregateOperationTimeoutError


    def flush(self):
        """
        Return the tetal number of rows deleted or inserted in this operation.
        (This is the number requested, in the case of deletes, not the number
        that actually matched.)

        This can be treated as a boolean to discover if anything was flushed.
        """
        count = 0
        if self.deletes:
            count += self._do_deletes()
            self.total_rows_deleted += count
            self.deletes.clear()
        if self.inserts:
            count += self._do_inserts()
            self.total_rows_inserted += count
            self.inserts.clear()
        self.total_size_inserted += self.size_added

        self.rows_added = 0
        self.size_added = 0
        self.bind_params_added = 0
        return count

    def _do_deletes(self):
        return self._do_batch('DELETE', sorted(iteritems(self.deletes)))

    def _do_batch(self, command, descriptors, rows_need_flattened=True, suffix=''):
        count = 0
        for (table, columns), rows in descriptors:
            count += len(rows)
            # Be careful not to use % formatting on a string already
            # containing `delete_placeholder`
            these_params_need_flattened = rows_need_flattened
            if len(columns) == 1:
                stmt, params, these_params_need_flattened = self._make_single_column_query(
                    command, table,
                    columns[0], rows, rows_need_flattened
                )
            else:
                row_template = " AND ".join(
                    ("%s=" % (column,)) + self.delete_placeholder
                    for column in columns
                )
                row_template = "(%s)" % (row_template,)
                rows_template = [row_template] * len(rows)
                stmt = "%s FROM %s WHERE %s" % (
                    command,
                    table, " OR ".join(rows_template))
                params = rows

            if these_params_need_flattened:
                params = self._flatten_params(params)
            stmt += suffix
            self.cursor.execute(stmt, params)

        return count

    def _flatten_params(self, params):
        params = sorted(params) if self.sorted_deletes else params
        params = list(itertools.chain.from_iterable(params))
        return params

    def _make_placeholder_list_of_length(self, count):
        return ','.join([self.delete_placeholder] * count)

    def _make_single_column_query(self, command, table,
                                  filter_column, filter_value,
                                  rows_need_flattened):
        placeholder_str = self._make_placeholder_list_of_length(len(filter_value))
        if not command.startswith('UPDATE'):
            stmt = "%s FROM %s WHERE %s IN (%s)" % (
                command, table, filter_column, placeholder_str
            )
        else:
            stmt = "%s WHERE %s IN (%s)" % (
                command, filter_column, placeholder_str
            )
        return stmt, filter_value, rows_need_flattened

    def _do_inserts(self):
        count = 0
        # In the case that we have only a single table we're inserting into,
        # and thus common key values, we don't need to sort or iterate.
        # If we have multiple tables, we never want to sort by the rows too,
        # that doesn't make any sense.
        if len(self.inserts) == 1:
            items = [self.inserts.popitem()]
        else:
            items = sorted(iteritems(self.inserts), key=lambda i: i[0])
        for (command, header, row_schema, suffix), rows in items:
            # Batched inserts
            rows = list(rows.values())
            count += len(rows)
            value_template = "(%s)" % row_schema
            values_template = [value_template] * len(rows)
            params = list(itertools.chain.from_iterable(rows))

            stmt = "%s INTO %s VALUES\n%s\n%s" % (
                command, header, ', '.join(values_template), suffix)
            # e.g.,
            # INSERT INTO table(c1, c2)
            # VALUES (%s, %s), (%s, %s), (%s, %s)
            # <suffix>
            __traceback_info__ = stmt
            self.cursor.execute(stmt, params)
        return count



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

import functools
import gc
import os
import tempfile
import unittest
from contextlib import contextmanager
from textwrap import dedent

import transaction

from zope.testing.setupstack import rmtree
from ZODB.DB import DB
from ZODB.FileStorage import FileStorage
from ZODB.blob import Blob
from ZODB.tests.util import TestCase

from relstorage.zodbconvert import main


def skipIfZapNotSupportedByDest(func):
    @functools.wraps(func)
    def test(self):
        if not self.zap_supported_by_dest:
            raise unittest.SkipTest("zap_all not supported")
        func(self)
    return test

class AbstractZODBConvertBase(TestCase):
    cfgfile = None

    # Set to True in a subclass if the destination can be zapped
    zap_supported_by_dest = False

    def setUp(self):
        super(AbstractZODBConvertBase, self).setUp()
        self._to_close = []
        self.__src_db = None
        self.__dest_db = None
        # zodb.tests.util.TestCase takes us to a temporary directory
        # as our working directory, and also sets it to tempfile.tempdir;
        # but don't require subclasses to know that.
        self.parent_temp_directory = tempfile.tempdir

        from relstorage.storage import copy
        self._orig_should_minor = copy._ProgressLogger._should_minor_log
        self._orig_should_major = copy._ProgressLogger._should_major_log
        self._orig_debug_enabled = copy._ProgressLogger.debug_enabled
        copy._ProgressLogger.debug_enabled = True
        copy._ProgressLogger._should_minor_log = lambda *args: True
        copy._ProgressLogger._should_major_log = lambda *args: True

    def tearDown(self):
        from relstorage.storage import copy
        copy._ProgressLogger._should_minor_log = self._orig_should_minor
        copy._ProgressLogger._should_major_log = self._orig_should_major
        copy._ProgressLogger.debug_enabled = self._orig_debug_enabled
        self.flush_changes_before_zodbconvert()
        self.close_all_registered()
        # XXX: On PyPy with psycopg2cffi, running these two tests will
        # result in a hang:
        # HPPostgreSQLDestZODBConvertTests.test_clear_empty_dest
        # HPPostgreSQLDestZODBConvertTests.test_clear_full_dest
        # test_clear_full_dest will hang in the zodbconvert call to
        # zap_all(), in the C code of the PG driver. Presumably some
        # connection with some lock got left open and was preventing
        # the TRUNCATE statements from taking out a lock. The same
        # tests do not hang with psycopg2cffi on C Python. Manually
        # running the gc (twice!) here fixes the issue. Note that this
        # only started when we wrapped the destination storage in
        # ZlibStorage (which copies methods into its own dict) so
        # there's something weird going on with the GC. Seen in PyPy
        # 2.5.0 and 5.3.
        gc.collect()
        gc.collect()
        super(AbstractZODBConvertBase, self).tearDown()

    def close_all_registered(self):
        for i in self._to_close:
            i.close()
        del self._to_close[:]

    src_db_needs_closed_before_zodbconvert = True
    dest_db_needs_closed_before_zodbconvert = True

    def flush_changes_before_zodbconvert(self):
        for db_name in 'src_db', 'dest_db':
            db_attr = '_' + AbstractZODBConvertBase.__name__ + '__' + db_name
            db = getattr(self, db_attr)
            needs_closed = getattr(self, db_name + '_needs_closed_before_zodbconvert')

            if needs_closed and db is not None:
                db.close()
                setattr(self, db_attr, None)
                if db in self._to_close:
                    self._to_close.remove(db)


    def _closing(self, thing):
        self._to_close.append(thing)
        return thing

    def _create_src_storage(self):
        raise NotImplementedError()

    def _create_dest_storage(self):
        raise NotImplementedError()

    def _create_src_db(self):
        if self.__src_db is None:
            self.__src_db = self._closing(DB(self._create_src_storage()))
        return self.__src_db

    def _create_dest_db(self):
        if self.__dest_db is None:
            self.__dest_db = self._closing(DB(self._create_dest_storage()))
        return self.__dest_db

    def _src_storage_from_db(self):
        # Do NOT close the returned storage
        return self._create_src_db().storage

    def _dest_storage_from_db(self):
        # Do NOT close the returned storage
        return self._create_dest_db().storage

    @contextmanager
    def __conn(self, name):
        db = getattr(self, '_create_' + name + '_db')()
        conn = db.open()
        try:
            yield conn
        finally:
            conn.close()

    def _src_conn(self):
        return self.__conn('src')

    def _dest_conn(self):
        return self.__conn('dest')

    def __write_value_for_key_in_db(self, val, key, db_conn_func):
        with db_conn_func() as conn:
            conn.root()[key] = val
            transaction.commit()

    def __check_value_of_key_in_db(self, val, key, db_conn_func):
        with db_conn_func() as conn2:
            db_val = conn2.root().get(key)
            if isinstance(val, Blob):
                self.assertIsInstance(db_val, Blob)
                with val.open('r') as f:
                    val = f.read()
                with db_val.open('r') as f:
                    db_val = f.read()

            self.assertEqual(db_val, val)

    def _write_value_for_key_in_src(self, x, key='x'):
        self.__write_value_for_key_in_db(x, key, self._src_conn)

    def _write_value_for_key_in_dest(self, x, key='x'):
        self.__write_value_for_key_in_db(x, key, self._dest_conn)

    def _check_value_of_key_in_dest(self, x, key='x'):
        self.__check_value_of_key_in_db(x, key, self._dest_conn)

    def _check_value_of_key_in_src(self, x, key='x'):
        self.__check_value_of_key_in_db(x, key, self._src_conn)

    def run_zodbconvert(self, *args):
        self.flush_changes_before_zodbconvert()
        return main(*args)

    def _storage_from_connection(self, conn):
        # Use a connection to get a storage so that you can be sure
        # it's been through ``new_instance``, etc.
        storage = conn._storage
        if not hasattr(storage, 'record_iternext'):
            # MVCCAdapter
            storage = storage._storage
        return storage

    def _flatten_db_iternext_storage(self, storage):
        # sync() to make sure the view is current, regardless of
        # where the storage came from.
        try:
            s = storage.sync
        except AttributeError:
            pass
        else:
            s(force=True)
        # [(oid, tid, state)]
        result = []
        cookie = None
        while 1:
            oid, tid, state, cookie = storage.record_iternext(cookie)
            result.append((oid, tid, state))
            if cookie is None:
                break
        # Iteration order is not guaranteed, although at this writing
        # both FileStorage and RelStorage iterate by OID
        result.sort()
        return result

    def deep_compare_current_states(self):
        src_state = self._flatten_db_iternext_storage(self._src_storage_from_db())
        dest_state = self._flatten_db_iternext_storage(self._src_storage_from_db())
        # The should both have something
        self.assertTrue(src_state)
        self.assertTrue(dest_state)
        # And they should all be equal
        self.assertEqual(src_state, dest_state)

        with self._src_conn() as c:
            src_from_conn = self._flatten_db_iternext_storage(
                self._storage_from_connection(c)
            )
        with self._dest_conn() as c:
            dest_from_conn = self._flatten_db_iternext_storage(
                self._storage_from_connection(c)
            )
        self.assertEqual(src_from_conn, src_state)
        self.assertEqual(dest_from_conn, dest_state)

    def test_convert(self):
        self._write_value_for_key_in_src(10)
        self.run_zodbconvert(['', self.cfgfile])
        self._check_value_of_key_in_dest(10)
        self.deep_compare_current_states()

    def test_dry_run(self):
        self._write_value_for_key_in_src(10)
        self.run_zodbconvert(['', '--dry-run', self.cfgfile])
        self._check_value_of_key_in_dest(None)

    def test_incremental_after_full_copy(self, first_convert_args=('',)):
        from persistent.mapping import PersistentMapping
        # Make sure to write new OIDs as well as new TIDs.
        for i in range(9):
            # Make sure to write new OIDs as well as new TIDs.
            self._write_value_for_key_in_src(PersistentMapping(i=i), key=i)
        self._write_value_for_key_in_src(Blob(b'abc'), 'the_blob')
        self._write_value_for_key_in_src(Blob(b'def'), 'the_blob')
        self._write_value_for_key_in_src(10)
        self._check_value_of_key_in_src(PersistentMapping(i=2), 2)

        self.run_zodbconvert(first_convert_args + (self.cfgfile,))

        self._check_value_of_key_in_dest(10)
        self._check_value_of_key_in_dest(PersistentMapping(i=2), 2)
        self._check_value_of_key_in_dest(Blob(b'def'), 'the_blob')
        self.deep_compare_current_states()

        for i in reversed(range(9)):
            # Make sure to write new OIDs as well as new TIDs,
            # But this time in *reverse* of the order we wrote them before.
            self._write_value_for_key_in_src(PersistentMapping(i=i**2), key=i)
        self._write_value_for_key_in_src(Blob(b'123'), 'the_blob')
        self._check_value_of_key_in_src(PersistentMapping(i=4), 2)
        self._write_value_for_key_in_src("hi")

        self.run_zodbconvert(['', '--incremental', self.cfgfile])

        self._check_value_of_key_in_dest("hi")
        self._check_value_of_key_in_dest(PersistentMapping(i=4), 2)
        self._check_value_of_key_in_dest(Blob(b'123'), 'the_blob')
        self.deep_compare_current_states()

    def test_incremental_after_incremental(self):
        # In case they do the conversion in different orders
        # depending on whether --incremental is specified or not.
        self.test_incremental_after_full_copy(first_convert_args=('', '--incremental'))

    def test_incremental_empty_src_dest(self):
        # Should work and not raise a POSKeyError
        self.run_zodbconvert(['', '--incremental', self.cfgfile])
        self._check_value_of_key_in_dest(None)

    @skipIfZapNotSupportedByDest
    def test_clear_empty_dest(self):
        x = 10
        self._write_value_for_key_in_src(x)
        self.run_zodbconvert(['', '--clear', self.cfgfile])
        self._check_value_of_key_in_dest(x)

    @skipIfZapNotSupportedByDest
    def test_clear_full_dest(self):
        self._write_value_for_key_in_dest(999)
        self._write_value_for_key_in_dest(666, key='y')
        self._write_value_for_key_in_dest(8675309, key='z')

        self._write_value_for_key_in_src(1, key='x')
        self._write_value_for_key_in_src(2, key='y')
        # omit z

        self.run_zodbconvert(['', '--clear', self.cfgfile])

        self._check_value_of_key_in_dest(1, key='x')
        self._check_value_of_key_in_dest(2, key='y')
        self._check_value_of_key_in_dest(None, key='z')

    def test_no_overwrite(self):
        self._create_src_db() # create the root object
        self._create_dest_db() # create the root object

        with self.assertRaises(SystemExit) as exc:
            self.run_zodbconvert(['', self.cfgfile])

        self.assertIn('Try --clear', str(exc.exception))

    def test_clear_dry_run(self):
        with self.assertRaises(SystemExit) as exc:
            self.run_zodbconvert(['', '--dry-run', '--clear', self.cfgfile])
        self.assertIn('Cannot clear a storage during a dry-run', str(exc.exception))

    # Also:
    # - when the destination keeps history: Verifying iteration
    #   of all the records, including transaction metadata
    # - verifying multiple states are saved when both keeps history
    # - If destination doesn't keep history, verifying only the most recent state# is saved.
    #
    # Some of that is probably handled in the History[Free|Preserving][From|To]FileStorageTest,
    # but I think that directly uses copyTransactionsFrom and doesn't have the other handling
    # zodbconvert does (e.g., incremental)


class FSZODBConvertTests(AbstractZODBConvertBase):
    keep_history = True

    def setUp(self):
        super(FSZODBConvertTests, self).setUp()

        fd, self.srcfile = tempfile.mkstemp('.fssource')
        os.close(fd)
        os.remove(self.srcfile)
        self.src_blobs = tempfile.mkdtemp('.fssource')

        fd, self.destfile = tempfile.mkstemp('.fsdest')
        os.close(fd)
        os.remove(self.destfile)
        self.dest_blobs = tempfile.mkdtemp('.fsdest')

        cfg = self._cfg_header() + self._cfg_source() + self._cfg_dest()
        self.cfgfile = self._write_cfg(cfg)

    def _cfg_header(self):
        return ""

    def _cfg_filestorage(self, name, path, blob_dir):
        return dedent("""
        <filestorage %s>
            path %s
            blob-dir %s
        </filestorage>
        """ % (name, path, blob_dir))

    def _cfg_one(self, name, path, blob_dir):
        return self._cfg_filestorage(name, path, blob_dir)

    def _cfg_source(self):
        return self._cfg_one('source', self.srcfile, self.src_blobs)

    def _cfg_dest(self):
        return self._cfg_one('destination', self.destfile, self.dest_blobs)

    def _write_cfg(self, cfg):
        fd, cfgfile = tempfile.mkstemp('.conf', 'zodbconvert-')
        os.write(fd, cfg.encode('ascii'))
        os.close(fd)
        return cfgfile

    def tearDown(self):
        # Close files first so they can be removed on Windows.
        self.flush_changes_before_zodbconvert()
        self.close_all_registered()

        def _rm(fname, meth):
            try:
                meth(fname)
            except OSError: # pragma: no cover
                import traceback
                print("Failed to remove", fname)
                traceback.print_exc()

        for fname in self.destfile, self.srcfile, self.cfgfile:
            if os.path.exists(fname):
                _rm(fname, os.remove)
        self.destfile = self.srcfile = self.cfgfile = None
        for dname in self.src_blobs, self.dest_blobs:
            if os.path.exists(dname):
                _rm(dname, rmtree)
        super(FSZODBConvertTests, self).tearDown()

    def _load_zconfig(self):
        from relstorage.zodbconvert import schema_xml
        from relstorage.zodbconvert import StringIO
        import ZConfig

        schema = ZConfig.loadSchemaFile(StringIO(schema_xml))
        conf, _ = ZConfig.loadConfig(schema, self.cfgfile)
        return conf

    def _create_src_storage(self):
        conf = self._load_zconfig()
        return conf.source.open()

    def _create_dest_storage(self):
        conf = self._load_zconfig()
        return conf.destination.open()

    def test_storage_has_data(self):
        from relstorage.zodbconvert import storage_has_data
        src = FileStorage(self.srcfile, create=True)
        self.assertFalse(storage_has_data(src))
        db = DB(src)  # add the root object
        db.close()
        self.assertTrue(storage_has_data(src))

class ZlibWrappedFSZODBConvertTests(FSZODBConvertTests):

    def _cfg_header(self):
        return "%import zc.zlibstorage\n"

    def _cfg_source(self):
        return ("\n<zlibstorage source>"
                + super(ZlibWrappedFSZODBConvertTests, self)._cfg_source()
                + "</zlibstorage>")

    def _cfg_dest(self):
        return ("\n<zlibstorage destination>"
                + super(ZlibWrappedFSZODBConvertTests, self)._cfg_dest()
                + "</zlibstorage>")

    def deep_compare_current_states(self):
        super(ZlibWrappedFSZODBConvertTests, self).deep_compare_current_states()
        from zc.zlibstorage import compress

        # Verify that both source and destination were compressed.
        src_raw = self._src_storage_from_db().base
        dest_raw = self._dest_storage_from_db().base

        src_states = self._flatten_db_iternext_storage(src_raw)
        dest_states = self._flatten_db_iternext_storage(dest_raw)
        for states in src_states, dest_states:
            __traceback_info__ = 'src' if states is src_states else 'dest'
            for _oid, _tid, state in states:
                __traceback_info__ = (__traceback_info__, _oid,)
                compressed = compress(state)
                self.assertEqual(compressed, state)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FSZODBConvertTests))
    suite.addTest(unittest.makeSuite(ZlibWrappedFSZODBConvertTests))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# -*- coding: utf-8 -*-
"""
Tests for vote.py.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import unittest

from hamcrest import assert_that
from nti.testing.matchers import verifiably_provides

from relstorage.tests import MockAdapter as _MockAdapter
from ...interfaces import ITPCStateVoting
from .. import SharedTPCState


class MockConnection(object):

    in_critical_phase = False
    exited_critical_phase = False
    entered_critical_phase = False
    cursor = None

    def exit_critical_phase(self):
        self.exited_critical_phase = True

    def enter_critical_phase_until_transaction_end(self):
        self.entered_critical_phase = True

    def rollback_quietly(self):
        """Does nothing"""

class MockStoreConnectionPool(object):

    def __init__(self):
        self.conn = MockConnection()

    def borrow(self):
        return self.conn

    def replace(self, conn):
        assert conn is self.conn

class MockTxnControl(object):
    def commit_phase1(self, _store_conn, _tid_int):
        """Does nothing."""

    def commit_phase2(self, _store_conn, _txn, _load_conn):
        """Does nothing"""

class MockAdapter(_MockAdapter):
    DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = True

    txncontrol = MockTxnControl()

    def lock_database_and_move(self, *_args, **_kwargs):
        """Does nothing"""
        return 12345678, None


class MockCache(object):
    def after_tpc_finish(self, *_args):
        """does nothing"""


class MockStorage(object):

    def __init__(self):
        self._store_connection_pool = MockStoreConnectionPool()
        self._adapter = MockAdapter()
        self._load_connection = MockConnection()
        self._cache = MockCache()

class MockBeginState(object):

    def __init__(self, shared_state=None):
        self.shared_state = shared_state or SharedTPCState(None, MockStorage(), None)
        self.required_tids = {}
        self.ude = (b'user', b'description', b'extension')
        self.invalidated_oids = set()

class MockLockedBeginState(MockBeginState):

    def __init__(self, shared_state=None):
        from ..vote import DatabaseLockedForTid
        MockBeginState.__init__(self, shared_state)
        self.committing_tid_lock = DatabaseLockedForTid(
            b'12345678',
            12345678,
            self.shared_state.adapter
        )

class _InterfaceMixin(object):

    BeginState = MockBeginState

    def setUp(self):
        from perfmetrics import set_statsd_client
        from perfmetrics import statsd_client
        from perfmetrics.testing import FakeStatsDClient
        from .. import vote
        self.stat_client = FakeStatsDClient()
        self.__orig_client = statsd_client()
        self.__orig_sample_rate = vote.METRIC_SAMPLE_RATE
        vote.METRIC_SAMPLE_RATE = 1
        set_statsd_client(self.stat_client)

    def tearDown(self):
        from perfmetrics import set_statsd_client
        from .. import vote
        set_statsd_client(self.__orig_client)
        vote.METRIC_SAMPLE_RATE = self.__orig_sample_rate

    def _getClass(self):
        raise NotImplementedError

    def _makeOne(self):
        return self._getClass()(self.BeginState())

    def test_provides_interface(self):
        assert_that(self._makeOne(), verifiably_provides(ITPCStateVoting))

    def _check_lock_and_move_commit(self, committed):
        self.assertTrue(committed)

    def test_exits_critical_section(self):
        vote = self._makeOne()
        vote.shared_state.adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = False
        assert not vote.shared_state.load_connection.exited_critical_phase
        assert not vote.shared_state.store_connection.exited_critical_phase
        committed = vote._lock_and_move()
        self._check_lock_and_move_commit(committed)
        if committed:
            self.assertTrue(vote.shared_state.load_connection.exited_critical_phase)
            self.assertTrue(vote.shared_state.store_connection.exited_critical_phase)
        else:
            self.assertFalse(vote.shared_state.load_connection.exited_critical_phase)
            self.assertFalse(vote.shared_state.store_connection.exited_critical_phase)

    def test_not_exits_critical_section_vote_only(self):
        vote = self._makeOne()
        vote.shared_state.adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = False
        assert not vote.shared_state.load_connection.exited_critical_phase
        assert not vote.shared_state.store_connection.exited_critical_phase
        committed = vote._lock_and_move(vote_only=True)
        self.assertFalse(committed)
        self.assertFalse(vote.shared_state.load_connection.exited_critical_phase)
        self.assertFalse(vote.shared_state.store_connection.exited_critical_phase)
        self.assertTrue(vote.shared_state.load_connection.entered_critical_phase)
        self.assertTrue(vote.shared_state.store_connection.entered_critical_phase)

    def test_not_exits_critical_section_interleavable(self):
        vote = self._makeOne()
        vote.shared_state.adapter.DEFAULT_LOCK_OBJECTS_AND_DETECT_CONFLICTS_INTERLEAVABLE = True
        assert not vote.shared_state.load_connection.exited_critical_phase
        assert not vote.shared_state.store_connection.exited_critical_phase
        committed = vote._lock_and_move()
        self._check_lock_and_move_commit(committed)
        self.assertFalse(vote.shared_state.load_connection.exited_critical_phase)
        self.assertFalse(vote.shared_state.store_connection.exited_critical_phase)

class _TPCFinishMixin(object):

    def test_tpc_finish_stats(self):
        from perfmetrics.testing.matchers import is_timer
        from hamcrest import contains_exactly
        vote = self._makeOne()
        vote.shared_state.transaction = self
        class InitialState(object):
            def with_committed_tid_int(self, *args):
                """Does nothing"""
        vote.shared_state.initial_state = InitialState()
        vote.lock_and_vote_times[0] = 1
        vote.lock_and_vote_times[1] = 1

        finish_times = [2, 3]
        def time():
            return finish_times.pop()

        vote.tpc_finish(vote.shared_state._storage, vote.transaction, _time=time)

        assert_that(self.stat_client,
                    contains_exactly(
                        # 2 - 1 as ms = 1000
                        is_timer('relstorage.storage.tpc_vote.objects_locked.t', '1000'),
                        # 3 - 1 as ms = 2000
                        is_timer('relstorage.storage.tpc_vote.between_vote_and_finish.t', '2000')
                    ))


class TestHistoryFree(_TPCFinishMixin, _InterfaceMixin, unittest.TestCase):

    def _getClass(self):
        from ..vote import HistoryFree
        return HistoryFree

class TestHistoryPreserving(_TPCFinishMixin, _InterfaceMixin, unittest.TestCase):

    BeginState = MockLockedBeginState

    def _getClass(self):
        from ..vote import HistoryPreserving
        return HistoryPreserving

class TestHistoryPreservingDeleteOnly(_InterfaceMixin, unittest.TestCase):

    BeginState = MockLockedBeginState

    def _getClass(self):
        from ..vote import HistoryPreservingDeleteOnly
        return HistoryPreservingDeleteOnly

    def _check_lock_and_move_commit(self, committed):
        self.assertFalse(committed)

class TestFunctions(unittest.TestCase):

    def _makeSharedState(self, initial_state=None, storage=None, transaction=None):
        return SharedTPCState(initial_state, storage, transaction)

    def _callFUT(self, ex, state, required_tids):
        from ..vote import add_details_to_lock_error
        add_details_to_lock_error(ex, state, required_tids)
        return ex

    def _check(self, kind):
        ex = kind('MESSAGE')
        state = self._makeSharedState()
        state.temp_storage.store_temp(1, b'abc', 42) # pylint:disable=no-member

        self._callFUT(ex, state, {1: 1})
        s = str(ex)
        self.assertIn('readCurrent {oid: tid}', s)
        self.assertIn('{1: 1}', s)
        self.assertIn('Previous TID', s)
        self.assertIn('42', s)
        self.assertIn('MESSAGE', s)

    def test_add_details_to_UnableToAcquireCommitLockError(self):
        from relstorage.adapters.interfaces import UnableToAcquireCommitLockError
        self._check(UnableToAcquireCommitLockError)

    def test_add_details_to_UnableToLockRowsToModifyError(self):
        from relstorage.adapters.interfaces import UnableToLockRowsToModifyError
        self._check(UnableToLockRowsToModifyError)

    def test_add_details_to_UnableToLockRowsToReadCurrentError(self):
        from relstorage.adapters.interfaces import UnableToLockRowsToReadCurrentError
        self._check(UnableToLockRowsToReadCurrentError)
